from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator
from enum import IntEnum
from pathlib import Path

from anyio import Event


class Change(IntEnum):
    added = 1
    modified = 2
    deleted = 3


FileChange = tuple[Change, str]


class FileWatcher(ABC):
    @abstractmethod
    async def watch(
        self,
        path: Path | str,
        stop_event: Event | None = None,
    ) -> AsyncGenerator[set[FileChange], None]: ...



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import AsyncExitStack
from inspect import isawaitable
from typing import TYPE_CHECKING, cast

from anyio import TASK_STATUS_IGNORED, Event, create_task_group
from anyio.abc import TaskGroup, TaskStatus

if TYPE_CHECKING:
    from pycrdt import Doc


class YStore(ABC):
    metadata_callback: Callable[[], Awaitable[bytes] | bytes] | None = None
    version = 2
    _started: Event | None = None
    _starting: bool = False
    _task_group: TaskGroup | None = None

    @abstractmethod
    def __init__(
        self, path: str, metadata_callback: Callable[[], Awaitable[bytes] | bytes] | None = None
    ): ...

    @abstractmethod
    async def write(self, data: bytes) -> None: ...

    @abstractmethod
    async def read(self) -> AsyncIterator[tuple[bytes, bytes]]: ...

    @property
    def started(self) -> Event:
        if self._started is None:
            self._started = Event()
        return self._started

    async def __aenter__(self) -> "YStore":
        if self._task_group is not None:
            raise RuntimeError("YStore already running")

        async with AsyncExitStack() as exit_stack:
            tg = create_task_group()
            self._task_group = await exit_stack.enter_async_context(tg)
            self._exit_stack = exit_stack.pop_all()
            await tg.start(self.start)

        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        await self.stop()
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        """Start the store.

        Arguments:
            task_status: The status to set when the task has started.
        """
        if self._starting:
            return

        self._starting = True

        if self._task_group is not None:
            raise RuntimeError("YStore already running")

        self.started.set()
        self._starting = False
        task_status.started()

    async def stop(self) -> None:
        """Stop the store."""
        if self._task_group is None:
            raise RuntimeError("YStore not running")

        self._task_group.cancel_scope.cancel()
        self._task_group = None

    async def get_metadata(self) -> bytes:
        """
        Returns:
            The metadata.
        """
        if self.metadata_callback is None:
            return b""

        metadata = self.metadata_callback()
        if isawaitable(metadata):
            metadata = await metadata
        metadata = cast(bytes, metadata)
        return metadata

    async def encode_state_as_update(self, ydoc: "Doc") -> None:
        """Store a YDoc state.

        Arguments:
            ydoc: The YDoc from which to store the state.
        """
        update = ydoc.get_update()
        await self.write(update)

    async def apply_updates(self, ydoc: "Doc") -> None:
        """Apply all stored updates to the YDoc.

        Arguments:
            ydoc: The YDoc on which to apply the updates.
        """
        async for update, *rest in self.read():  # type: ignore
            ydoc.apply_update(update)


class YDocNotFound(Exception):
    pass


class YStoreFactory:
    def __init__(self, ystore_factory: type[YStore]) -> None:
        self._ystore_factory = ystore_factory

    def __call__(self, *args, **kwargs) -> YStore:
        return self._ystore_factory(*args, **kwargs)



from abc import ABC, abstractmethod

from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import HTMLResponse
from fps import Context, put

from jupyverse_api import Config, Router

from ..app import App
from ..auth import Auth, User


class JupyterLab(Router, ABC):
    def __init__(self, app: App, auth: Auth):
        super().__init__(app)

        router = APIRouter()

        @router.get("/lab")
        async def get_lab(
            request: Request,
            user: User = Depends(auth.current_user()),
        ):
            async with Context():
                put(request)
                return await self.get_lab("lab", user)

        @router.get("/doc")
        async def get_doc(
            request: Request,
            user: User = Depends(auth.current_user()),
        ):
            async with Context():
                put(request)
                return await self.get_lab("doc", user)

        @router.get("/{mode}/tree/{path:path}")
        async def load_workspace(
            request: Request,
            mode,
            path,
        ):
            if mode not in {"lab", "doc"}:
                raise HTTPException(status_code=404, detail="Not found")

            async with Context():
                put(request)
                return await self.load_workspace(mode, path)

        @router.get("/lab/api/workspaces/{name}")
        @router.get("/doc/api/workspaces/{name}")
        async def get_workspace_data(
            request: Request,
            user: User = Depends(auth.current_user()),
        ):
            async with Context():
                put(request)
                return await self.get_workspace_data(user)

        @router.put(
            "/lab/api/workspaces/{name}",
            status_code=204,
        )
        @router.put(
            "/doc/api/workspaces/{name}",
            status_code=204,
        )
        async def set_workspace(
            request: Request,
            user: User = Depends(auth.current_user()),
            user_update=Depends(auth.update_user),
        ):
            return await self.set_workspace(request, user, user_update)

        @router.get("/{mode}/workspaces/{name}", response_class=HTMLResponse)
        async def get_workspace(
            request: Request,
            mode,
            name,
            user: User = Depends(auth.current_user()),
        ):
            if mode not in {"lab", "doc"}:
                raise HTTPException(status_code=404, detail="Not found")

            async with Context():
                put(request)
                return await self.get_workspace(mode, name, "", user)

        @router.get("/{mode}/workspaces/{name}/tree/{path:path}", response_class=HTMLResponse)
        async def get_workspace_with_tree(
            request: Request,
            mode,
            name,
            path,
            user: User = Depends(auth.current_user()),
        ):
            if mode not in {"lab", "doc"}:
                raise HTTPException(status_code=404, detail="Not found")

            async with Context():
                put(request)
                return await self.get_workspace(mode, name, path, user)

        self.include_router(router)

    @abstractmethod
    async def get_lab(self, mode, user: User): ...

    @abstractmethod
    async def load_workspace(self, mode, path): ...

    @abstractmethod
    async def get_workspace_data(self, user: User): ...

    @abstractmethod
    async def set_workspace(
        self,
        request: Request,
        user: User,
        user_update,
    ): ...

    @abstractmethod
    async def get_workspace(self, mode, name, path, user: User): ...


class JupyterLabConfig(Config):
    dev_mode: bool = False
    server_side_execution: bool = False



from collections.abc import Awaitable, Callable

from fps import Module


class PageConfig:
    def __init__(self) -> None:
        self._config: dict[str, str] = {}
        self._hooks: list[Callable[[dict[str, str]], Awaitable[None]]] = []

    def register(self, hook: Callable[[dict[str, str]], Awaitable[None]]):
        self._hooks.append(hook)

    def set(self, **kwargs: str) -> None:
        self._config = kwargs

    async def get(self) -> dict[str, str]:
        for hook in self._hooks:
            await hook(self._config)

        return self._config


class PageConfigModule(Module):
    async def prepare(self):
        self.put(PageConfig())



from pydantic import BaseModel


class Checkpoint(BaseModel):
    id: str
    last_modified: str


class Content(BaseModel):
    name: str
    path: str
    last_modified: str | None = None
    created: str | None = None
    content: list[dict] | str | dict | None = None
    format: str | None = None
    mimetype: str | None = None
    size: int | None = None
    writable: bool
    type: str


class CreateContent(BaseModel):
    ext: str | None = None
    path: str
    type: str


class SaveContent(BaseModel):
    content: str | dict | None = None
    format: str
    path: str
    type: str


class RenameContent(BaseModel):
    path: str



import io
import json
from importlib.metadata import entry_points
from typing import Any

import rich_click as click
from fps.cli._cli import main as fps_main


@click.command()  # type: ignore
@click.option(
    "--debug",
    is_flag=True,
    show_default=True,
    default=False,
    help="Enable debug mode.",
)
@click.option(
    "--show-config",
    is_flag=True,
    show_default=True,
    default=False,
    help="Show the actual configuration.",
)
@click.option(
    "--help-all",
    is_flag=True,
    show_default=True,
    default=False,
    help="Show the configuration description.",
)
@click.option(
    "--backend",
    show_default=True,
    default="asyncio",
    help="The name of the event loop to use (asyncio or trio).",
)
@click.option(
    "--open-browser",
    is_flag=True,
    show_default=True,
    default=False,
    help="Open a browser window.",
)
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="The host URL.",
)
@click.option(
    "--port",
    type=int,
    default=8000,
    help="The host port.",
)
@click.option(
    "--query-param",
    multiple=True,
    type=str,
    help='The query parameter key and value, separated by "=".',
)
@click.option(
    "--allow-origin",
    multiple=True,
    type=str,
    help="The origin to allow.",
)
@click.option(
    "--set",
    "set_",
    multiple=True,
    type=str,
    help="Set configuration.",
)
@click.option(
    "--disable",
    multiple=True,
    type=str,
    help="Disable plugin.",
)
@click.option(
    "--timeout",
    type=float,
    default=None,
    help="The timeout for starting Jupyverse.",
)
@click.option(
    "--stop-timeout",
    type=float,
    default=1,
    help="The timeout for stopping Jupyverse.",
)
def main(
    debug: bool = False,
    show_config: bool = False,
    help_all: bool = False,
    backend: str = "asyncio",
    open_browser: bool = False,
    host: str = "127.0.0.1",
    port: int = 8000,
    set_: tuple[str, ...] = (),
    disable: tuple[str, ...] = (),
    allow_origin: tuple[str, ...] = (),
    query_param: tuple[str, ...] = (),
    timeout: float | None = None,
    stop_timeout: float = 1,
) -> None:
    query_params_dict = {}
    for qp in query_param:
        key, _, value = qp.partition("=")
        query_params_dict[key] = value
    query_params_str = json.dumps(query_params_dict)
    allow_origins_str = json.dumps(allow_origin)
    set_list: list[str] = list(set_)
    set_list.append(f"debug={debug}")
    set_list.append(f"open_browser={open_browser}")
    set_list.append(f"host={host}")
    set_list.append(f"port={port}")
    set_list.append(f"allow_origins={allow_origins_str}")
    set_list.append(f"query_params={query_params_str}")
    pluggin_config = io.StringIO(json.dumps(get_pluggin_config(disable)))
    fps_main.callback(
        "",
        set_=set_list,
        config=pluggin_config,
        show_config=show_config,
        help_all=help_all,
        backend=backend,
        timeout=timeout,
        stop_timeout=stop_timeout,
    )  # type: ignore


def get_pluggin_config(disable: tuple[str, ...]) -> dict[str, Any]:
    jupyverse_modules = [
        ep.name for ep in entry_points(group="jupyverse.modules") if ep.name not in disable
    ]
    config = {
        "jupyverse": {
            "type": "jupyverse_api.main:JupyverseModule",
            "modules": {module: {"type": module} for module in jupyverse_modules},
        }
    }
    return config



from abc import ABC, abstractmethod

from anyio import TASK_STATUS_IGNORED, Event, create_memory_object_stream
from anyio.abc import TaskStatus
from anyio.streams.memory import MemoryObjectReceiveStream
from anyio.streams.stapled import StapledObjectStream


class Kernel(ABC):
    def __init__(self) -> None:
        self.key = "0"
        self.wait_for_ready = False
        self.started = Event()

        self._to_shell_send_stream, self._to_shell_receive_stream = create_memory_object_stream[
            list[bytes]
        ]()
        self._from_shell_send_stream, self._from_shell_receive_stream = create_memory_object_stream[
            list[bytes]
        ]()
        self._to_control_send_stream, self._to_control_receive_stream = create_memory_object_stream[
            list[bytes]
        ]()
        self._from_control_send_stream, self._from_control_receive_stream = (
            create_memory_object_stream[list[bytes]]()
        )
        self._to_stdin_send_stream, self._to_stdin_receive_stream = create_memory_object_stream[
            list[bytes]
        ]()
        self._from_stdin_send_stream, self._from_stdin_receive_stream = create_memory_object_stream[
            list[bytes]
        ]()
        self._from_iopub_send_stream, self._from_iopub_receive_stream = create_memory_object_stream[
            list[bytes]
        ](max_buffer_size=float("inf"))
        self._shell_stream = StapledObjectStream(
            self._to_shell_send_stream, self._from_shell_receive_stream
        )
        self._control_stream = StapledObjectStream(
            self._to_control_send_stream, self._from_control_receive_stream
        )
        self._stdin_stream = StapledObjectStream(
            self._to_stdin_send_stream, self._from_stdin_receive_stream
        )

    @abstractmethod
    async def start(
        self,
        *,
        task_status: TaskStatus[None] = TASK_STATUS_IGNORED,
    ) -> None: ...

    @abstractmethod
    async def stop(self) -> None: ...

    @abstractmethod
    async def interrupt(self) -> None: ...

    @property
    def shell_stream(self) -> StapledObjectStream[list[bytes]]:
        return self._shell_stream

    @property
    def control_stream(self) -> StapledObjectStream[list[bytes]]:
        return self._control_stream

    @property
    def stdin_stream(self) -> StapledObjectStream[list[bytes]]:
        return self._stdin_stream

    @property
    def iopub_stream(self) -> MemoryObjectReceiveStream[list[bytes]]:
        return self._from_iopub_receive_stream


class KernelFactory:
    def __init__(self, kernel_factory: type[Kernel]) -> None:
        self._kernel_factory = kernel_factory

    def __call__(self, *args, **kwargs) -> Kernel:
        return self._kernel_factory(*args, **kwargs)


class DefaultKernelFactory(KernelFactory):
    pass
